import io
from io import BytesIO
from pathlib import Path

import boto3
import cv2
import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
import s3fs
from loguru import logger
from PIL import Image

s3 = boto3.client("s3")


def list_objects_in_folder(bucket_name, folder_path):
    # List objects in the specified folder
    response = s3.list_objects_v2(Bucket=bucket_name, Prefix=folder_path)

    # Print the object keys
    print(f"Objects in the folder '{folder_path}':")
    for obj in response.get("Contents", []):
        print(obj["Key"])


def display_s3_image(bucket_name, object_key):
    # Download the image from S3
    response = s3.get_object(Bucket=bucket_name, Key=object_key)
    image_data = response["Body"].read()

    # Open and display the image using Pillow (PIL)
    return Image.open(BytesIO(image_data))


def get_s3_image(
    img_name,
    bucket_name="geoml-aquarry",
    path="output/gms/16dec2023/tang",
    save_path=None,
):
    # Get the file object from S3
    object_key = f"{path}/{img_name}.jpg"
    try:
        file_object = s3.get_object(Bucket=bucket_name, Key=object_key)
        file_stream = io.BytesIO(file_object["Body"].read())
        img_np_array = np.asarray(bytearray(file_stream.read()), dtype=np.uint8)
        img = cv2.imdecode(img_np_array, cv2.IMREAD_COLOR)

        # Optionally save to disk
        if save_path:
            local_file_path = f"{save_path}/{img_name}.jpg"
            with open(local_file_path, "wb") as file:
                file.write(file_stream.getvalue())

        return img
    except Exception as e:
        logger.debug(
            "Error accessing object_key: {}, bucket_name: {}. Exception: {}",
            object_key,
            bucket_name,
            e,
        )
        return None


def write_to_parquet(data, objectid, path_output_file):
    path_output_file = str(path_output_file)
    # Prepare data for each column
    labels = [data["labels"][i].item() for i in range(len(data["labels"]))]
    scores = [data["scores"][i].item() for i in range(len(data["scores"]))]
    boxes_normalized = [
        data["boxes_normalized"][i] for i in range(len(data["boxes_normalized"]))
    ]
    objectid_arr = [str(objectid)] * len(data["labels"])
    try:
        # Create a PyArrow Table
        table = pa.Table.from_arrays(
            [
                pa.array(objectid_arr, pa.string()),
                pa.array(labels, pa.int64()),
                pa.array(scores, pa.float64()),
                pa.array(boxes_normalized, pa.list_(pa.float64())),
            ],
            names=["objectid", "label", "score", "boxes_normalized"],
        )

        # If starts with `s3://` write to S3, else to disk
        if path_output_file.startswith("s3://"):
            # Use S3FileSystem to handle S3 paths
            fs = s3fs.S3FileSystem()
            with fs.open(path_output_file, "wb") as f:
                pq.write_table(table, f)
        # Write the table to a Parquet file
        pq.write_table(table, path_output_file)
        logger.debug("Successfully wrote to: {}", path_output_file)
    except Exception as e:
        logger.error("Failed to write to: {} with {}", path_output_file, e)
        raise e


def structure_s3_key(model_name, dataset_name, img_transform, filename, s3base):
    return "s3://" + str(
        Path(s3base)
        / f"dataset={dataset_name}"
        / f"model={model_name}"
        / f"color={img_transform}"
        / f"{filename}.parquet"
    )



import glob as glob
import multiprocessing
import os
from pathlib import Path
from typing import Literal

import cv2
import numpy as np
import torch
import torchvision
from ensemble_boxes import weighted_boxes_fusion
from loguru import logger
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor

from geomlcv.utils.process_images.utils import _pcahog, _svhog

PATH_BASE = Path("/home/ec2-user/geomlcv/")  # TODO: from config
PATH_WEIGHTS = Path(
    os.getenv("PATH_WEIGHTS", PATH_BASE / "docs/notebooks/.data/weights")
)
MODELS = {
    "RCNN": {
        "rgb": PATH_WEIGHTS
        / "fasterrcnn_resnet50_fpnv2_pretrained_rgb_v5_640_2023-12-12.pth",
        "svhog": PATH_WEIGHTS
        / "fasterrcnn_resnet50_fpnv2_pretrained_svhog_v3_640_2023-12-13.pth",
        "pcahog": PATH_WEIGHTS
        / "fasterrcnn_resnet50_fpnv2_pretrained_pcahog_v3_640_2023-12-14.pth",
        "classes": ["__background__", "non-pitlake-water", "pitlake"],
    },
    "YOLO": {
        "rgb": "",
        "svhog": "",
        "pcahog": "",
        "classes": ["non-pitlake-water", "pitlake"],
    },
}


def create_RCNN_model(num_classes):
    model = torchvision.models.detection.fasterrcnn_resnet50_fpn_v2(
        weights="FasterRCNN_ResNet50_FPN_V2_Weights.DEFAULT"
    )
    in_features = model.roi_heads.box_predictor.cls_score.in_features
    model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes)

    return model


def fasterRCNN_inference(model, image):
    orig_image = image.copy()
    image = cv2.cvtColor(orig_image, cv2.COLOR_BGR2RGB).astype(np.float32)
    image /= 255.0
    image = np.transpose(image, (2, 0, 1)).astype(np.float32)
    if torch.cuda.is_available():
        image = torch.tensor(image, dtype=torch.float).cuda()
    else:
        image = torch.tensor(image, dtype=torch.float)
    image = torch.unsqueeze(image, 0)
    with torch.no_grad():
        outputs = model(image.to("cpu"))
    outputs = [{k: v.to("cpu") for k, v in t.items()} for t in outputs]

    return outputs


def create_YOLO_model(num_classes):
    raise


def YOLOv8_inference(model, image):
    raise


def select_model(model_name, model_path, classes):
    if model_name == "RCNN":
        model = create_RCNN_model(num_classes=len(classes))
        checkpoint = torch.load(model_path, map_location="cpu")
        model.load_state_dict(checkpoint["model_state_dict"])
        if torch.cuda.is_available():
            model.eval()
        else:
            model.to("cpu").eval()
        return model
    elif model_name == "YOLO":
        raise ("no model found")


def normalize_boxes(boxes, norm):
    for k in range(boxes.shape[0]):
        for i in range(4):
            boxes[k][i] = boxes[k][i] / norm[i]

    return boxes


def run_inference(
    img,
    model_name: Literal["RCNN", "YOLO"],
    img_transform: Literal["rgb", "svhog", "pcahog"],
    confidence_threshold=0.5,
):
    # Preprocess input image with faux-color
    if img_transform == "svhog":
        img = _svhog(img)
    elif img_transform == "pcahog":
        img = _pcahog(img)

    # Inference
    if model_name == "RCNN":
        out = fasterRCNN_inference(
            select_model(
                model_name,
                MODELS[model_name][img_transform],
                MODELS[model_name]["classes"],
            ),
            img,
        )[0]
    else:
        raise ("Model not found.")

    # Normalize
    h, w, _ = img.shape
    out["boxes_normalized"] = normalize_boxes(out["boxes"].data.numpy(), [w, h, w, h])

    # Filter results based on the confidence threshold
    scores = out["scores"].data.numpy()
    keep_indices = scores >= confidence_threshold

    out["boxes"] = out["boxes"][keep_indices]
    out["labels"] = out["labels"][keep_indices]
    out["scores"] = out["scores"][keep_indices]
    out["boxes_normalized"] = out["boxes_normalized"][keep_indices]

    return out


def _worker(img, model_name, img_transform):
    return run_inference(img, model_name=model_name, img_transform=img_transform)


def run_combined_inference(img, results_dir, image_name, detection_threshold=0.2):
    # Run inferences in parallel
    # Create a pool of processes
    with multiprocessing.Pool() as pool:
        # Map the run_inference function to multiple arguments
        results = pool.starmap(
            _worker,
            [(img, "RCNN", "rgb"), (img, "RCNN", "svhog"), (img, "RCNN", "pcahog")],
        )

    # Unpack results
    rgb_output, svhog_output, pcahog_output = results

    # TODO: cleanup with dict comprehension
    boxes_list = [
        rgb_output[0]["boxes_normalized"].data.numpy(),
        svhog_output[0]["boxes_normalized"].data.numpy(),
        pcahog_output[0]["boxes_normalized"].data.numpy(),
    ]
    scores_list = [
        rgb_output[0]["scores"].data.numpy(),
        svhog_output[0]["scores"].data.numpy(),
        pcahog_output[0]["scores"].data.numpy(),
    ]
    labels_list = [
        rgb_output[0]["labels"].data.numpy(),
        svhog_output[0]["labels"].data.numpy(),
        pcahog_output[0]["labels"].data.numpy(),
    ]

    # Hyperparameters for weighted fusion
    weights = [3, 2, 1]
    iou_thr = 0.8
    skip_box_thr = 0.0001
    detection_threshold = 0.2

    boxes, scores, labels = weighted_boxes_fusion(
        boxes_list,
        scores_list,
        labels_list,
        weights=weights,
        iou_thr=iou_thr,
        skip_box_thr=skip_box_thr,
    )

    # carry further only if there are detected boxes
    h, w, _ = img.shape
    if len(boxes) != 0:
        # filter out boxes according to `detection_threshold`
        _indices = scores >= detection_threshold
        boxes_filtered = boxes[_indices]
        scores_filtered = scores[_indices]
        labels_filtered = labels[_indices].astype(int)
        draw_boxes = boxes_filtered.copy()
        classes = MODELS["RCNN"]["classes"]

        # TODO: write these out to a file, perhaps separately prior to weighted
        # NOTE: rcnn has 3 classes, bc background.
        # TODO: this maps class int to class name

        # get all the predicited class names
        pred_classes_filtered = [classes[i] for i in labels_filtered]

        # draw the bounding boxes and write the class name on top of it
        for j, box in enumerate(draw_boxes):
            class_name = pred_classes_filtered[j]
            score_j = scores_filtered[j]
            textToDisplay = class_name + ", score=" + str(np.round(score_j, 2))

            cv2.rectangle(
                img,
                (int(box[0] * w), int(box[1] * h)),
                (int(box[2] * w), int(box[3] * h)),
                (255, 255, 255),
                4,
            )
            cv2.putText(
                img,
                textToDisplay,
                (int(box[0] * w), int((box[1] * h) - 5)),
                cv2.FONT_HERSHEY_SIMPLEX,
                1,
                (0, 0, 255),
                4,
                lineType=cv2.LINE_AA,
            )

        output_file = Path(results_dir) / f"{image_name}_wf_inference.jpg"
        cv2.imwrite(str(output_file), img)
    logger.info("run_combined_inference completed successfully")
    return



import multiprocessing


def run_parallel(func, arr_args, num_cores=multiprocessing.cpu_count()):
    pool = multiprocessing.Pool(num_cores)
    data = pool.map(func, arr_args)
    pool.close()
    pool.join()
    return data



import base64
import hashlib
import hmac
import urllib.parse as urlparse

import requests

# Get signing secret from: https://developers.google.com/maps/documentation/streetview/digital-signature#sample-code-for-url-signing


def _sign_url(input_url, secret):
    """
    Sign a URL with a secret using HMAC SHA1.

    This function creates a signed URL that can be used with Google Maps services.
    It signs the path and query components of the URL.

    Args:
    input_url (str): The URL to be signed.
    secret (str): The URL-safe, base64-encoded secret key.

    Returns:
    str: The input URL with a signature parameter appended.
    """
    url = urlparse.urlparse(input_url)

    # We only need to sign the path+query part of the string
    url_to_sign = url.path + "?" + url.query

    # Decode the private key into its binary format
    # We need to decode the URL-encoded private key
    decoded_key = base64.urlsafe_b64decode(secret)

    # Create a signature using the private key and the URL-encoded
    # string using HMAC SHA1. This signature will be binary.
    signature = hmac.new(decoded_key, str.encode(url_to_sign), hashlib.sha1)

    # Encode the binary signature into base64 for use within a URL
    encoded_signature = base64.urlsafe_b64encode(signature.digest())

    original_url = url.scheme + "://" + url.netloc + url.path + "?" + url.query

    # Return signed URL
    return original_url + "&signature=" + encoded_signature.decode()


def _gms_static_map_signed_url(
    lat: str,
    lon: str,
    api_key: str,
    secret: str | None = None,
    size: int = 640,
    zoom: int = 16,
    maptype: str = "satellite",
    format: str = "jpg",
):
    """
    Generate a signed URL for Google Maps Static API.

    This function constructs a URL for fetching a static map image from Google Maps,
    and signs it if a secret key is provided.

    Args:
    lat (str): Latitude of the map's center.
    lon (str): Longitude of the map's center.
    api_key (str): Google Maps API key.
    secret (str, optional): Secret key for signing the URL.
    size (int, optional): Width and height of the map image in pixels. Default is 640.
    zoom (int, optional): Zoom level of the map. Default is 16.
    maptype (str, optional): Type of the map (e.g., "satellite", "roadmap").
    format (str, optional): Format of the map image (e.g., "jpg", "png").

    Returns:
    str: A URL to the static map image, signed if a secret is provided.
    """
    input_url = f"https://maps.googleapis.com/maps/api/staticmap?center={lat},{lon}&zoom={zoom}&size={size}x{size}&maptype={maptype}&format={format}&key={api_key}"
    if secret:
        return _sign_url(input_url, secret)
    else:
        return input_url


def _gms_static_map_request(signed_url: str):
    """
    Send a request to the Google Maps Static API.

    Fetches a static map image from Google Maps using a signed URL.

    Args:
    signed_url (str): The signed URL to fetch the static map.

    Returns:
    bytes: The content of the response, typically an image file.

    Raises:
    Exception: If the HTTP request fails.
    """
    response = requests.get(signed_url)
    if response.status_code == 200:
        return response.content
    else:
        raise Exception(f"Error: {response.status_code} - {response.text}")


def gms_static_map(
    lat: str,
    lon: str,
    api_key: str,
    secret: str | None = None,
    size: int = 640,
    zoom: int = 16,
    maptype: str = "satellite",
    format: str = "jpg",
):
    """
    Main function to obtain a static map from Google Maps Static API.

    Constructs and signs a URL, then fetches the static map image.

    Args:
    lat (str): Latitude of the map's center.
    lon (str): Longitude of the map's center.
    api_key (str): Google Maps API key.
    secret (str, optional): Secret key for signing the URL.
    size (int, optional): Width and height of the map image in pixels. Default is 640.
    zoom (int, optional): Zoom level of the map. Default is 16.
    maptype (str, optional): Type of the map (e.g., "satellite", "roadmap").
    format (str, optional): Format of the map image (e.g., "jpg", "png").

    Returns:
    bytes: The content of the response, typically an image file.

    Example:
    gms_static_map(lat = 10.714728, lon = -73.998672, api_key=api_key, secret=secret)
    """
    signed_url = _gms_static_map_signed_url(
        lat=lat,
        lon=lon,
        api_key=api_key,
        secret=secret,
        size=size,
        zoom=zoom,
        maptype=maptype,
        format=format,
    )
    return _gms_static_map_request(signed_url)
